# !/usr/bin/env python
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# Title       : car_state.py
# Author      : aDd1kTeD2Ka0s (add1kted2ka0s@gmail.com)
# Date        : 20-Jun-2014
# Description : Ka0s Leaderboard - Race Status
#             : Original code by Rombik
# -----------------------------------------------------------------------------

# -------------------------------------
# Imports
# -------------------------------------

import ac
import acsys
import itertools
import re

from .config import *
from ..lib.json_data import JsonData
from ..lib.sim_info import info
from ..lib.utils import Signal, log

# -------------------------------------
# Globals
# -------------------------------------
CAR_NAME_DICT = dict()
TRACK_LENGTH_DICT = dict()
RE_NON_DECIMAL = re.compile(r'[^\d.]+')

# -------------------------------------
class CarState:
	# -------------------------------------

	# -------------------------------------
	def __init__(self, car_id=0):
		# -------------------------------------
		self.counter = itertools.count()
		self.on_new_lap = Signal()

		self.car_id = car_id
		self.car_name_id = ac.getCarName(self.car_id)
		self.car_name = self.getFullCarName()
		self.track_name_id = ac.getTrackName(self.car_id)
		self.track_configuration = ac.getTrackConfiguration(self.car_id)
		self.driver_name = ac.getDriverName(self.car_id)
		self.number_of_laps = info.graphics.numberOfLaps if info.graphics.numberOfLaps > 0 else 2
		self.track_length_km, self.track_length_mi = self.getTrackLength()

		self.reset()

	# -------------------------------------
	def getFullCarName(self):
		# -------------------------------------
		if self.car_name_id in CAR_NAME_DICT:
			car_name = CAR_NAME_DICT[self.car_name_id]
		else:
			try:
				car_data = JsonData(type='cars', key=self.car_name_id)
				car_name = car_data.get('name')
				CAR_NAME_DICT[self.car_name_id] = car_name
			except Exception:
				car_name = self.car_name_id

		return car_name

	# -------------------------------------
	def getTrackLength(self):
		# -------------------------------------
		if 'km' in TRACK_LENGTH_DICT:
			track_length_km = TRACK_LENGTH_DICT['km']
			track_length_mi = TRACK_LENGTH_DICT['mi']
		else:
			try:
				track_data = JsonData(type='tracks', key=self.track_name_id)
				track_length_str = track_data.get('length')
				track_length_nondecimal = RE_NON_DECIMAL.sub('', track_length_str)
				track_length_km = float(track_length_nondecimal)
				track_length_mi = round(track_length_km / KM_TO_MILES, 3)
				TRACK_LENGTH_DICT['km'] = track_length_km
				TRACK_LENGTH_DICT['mi'] = track_length_mi
			except Exception:
				track_length_km = track_length_mi = 0

		return track_length_km, track_length_mi

	# -------------------------------------
	def reset(self):
		# -------------------------------------
		self._lap_handled = False
		self._last_lap_number = 0
		self._total_time = 0

		self.completed_laps = 0
		self.completion_lap = 0
		self.completion_race = 0
		self.distance_lap_km = 0
		self.distance_lap_mi = 0
		self.distance_race_km = 0
		self.distance_race_mi = 0
		self.last_splits = ac.getLastSplits(self.car_id)
		self.current_splits = ac.getCurrentSplits(self.car_id)
		self.live_splits = ac.getCurrentSplits(self.car_id)
		self.split_count = len(ac.getCurrentSplits(self.car_id))
		self.inprogress_split = 0
		self.first_line_crossed = False
		self.position = 0
		self.last_lap = 0
		self.best_lap = 0
		self.current_lap = 0
		self.speed_ms_current = 0
		self.speed_mph_current = 0
		self.speed_kmh_current = 0
		self.speed_ms_max = 0
		self.speed_mph_max = 0
		self.speed_kmh_max = 0
		self.laps = dict()
		self.splits = dict()

	# -------------------------------------
	def on_update(self, dt, inf=float('inf')):
		# -------------------------------------
		self.current_lap = ac.getCarState(self.car_id, acsys.CS.LapTime)
		if self.current_lap == -1:
			return

		self.position = ac.getCarState(self.car_id, acsys.CS.NormalizedSplinePosition)
		self.last_lap = sum(ac.getLastSplits(self.car_id))
		self.best_lap = ac.getCarState(self.car_id, acsys.CS.BestLap)
		self.completed_laps = ac.getCarState(self.car_id, acsys.CS.LapCount)
		self.number_of_laps = info.graphics.numberOfLaps if info.graphics.numberOfLaps > 0 else 1

		if self.first_line_crossed == False and self.completed_laps == 0 and self.position <= 0.01:
			self.first_line_crossed = True
		if self.first_line_crossed == True:
			self.completion_lap = self.position * 100.0
			self.completion_race = (self.completed_laps + self.position) / self.number_of_laps * 100.0
		else:
			self.completion_lap = (1 - self.position) * 100.0 * -1.0
			self.completion_race = (1 - self.position) / self.number_of_laps * 100 * -1.0

		self.distance_lap_km = self.completion_lap * self.track_length_km / 100.0
		self.distance_lap_mi = self.completion_lap * self.track_length_mi / 100.0
		self.distance_race_km = (self.completed_laps * self.track_length_km) + self.distance_lap_km
		self.distance_race_mi = (self.completed_laps * self.track_length_mi) + self.distance_lap_mi

		self.last_splits = ac.getLastSplits(self.car_id)
		self.current_splits = ac.getCurrentSplits(self.car_id)

		self.inprogress_split = -1
		for i, current_split in enumerate(self.current_splits):
			if current_split == 0:
				if self.inprogress_split == -1:
					self.inprogress_split = i
				if self.inprogress_split == i:
					self.live_splits[i] = (self.current_lap - sum(self.current_splits))
				else:
					if PREVIOUS_LAP_SPLITS_DECAY:
						# Use this code path if you want the previous lap splits to persist for 'X' seconds
						if self.current_lap < PREVIOUS_LAP_SPLITS_DECAY_TIME:
							self.live_splits[i] = self.last_splits[i]
						else:
							self.live_splits[i] = 0
					else:
						# Use this code path if you want the previous lap splits to persist until you enter that sector
						self.live_splits[i] = self.last_splits[i]
			else:
				self.live_splits[i] = current_split

		self.speed_mph_current = ac.getCarState(self.car_id, acsys.CS.SpeedMPH)
		if self.speed_mph_current > self.speed_mph_max:
			self.speed_mph_max = self.speed_mph_current

		self.speed_kmh_current = ac.getCarState(self.car_id, acsys.CS.SpeedKMH)
		if self.speed_kmh_current > self.speed_kmh_max:
			self.speed_kmh_max = self.speed_kmh_current

		self.laps[self.completed_laps + 1] = self.current_lap
		self.splits[self.completed_laps + 1] = self.current_splits
		for i, current_split in enumerate(self.current_splits):
			if current_split == 0:
				if self.inprogress_split == i:
					self.splits[self.completed_laps + 1][i] = (self.current_lap - sum(self.current_splits))
				else:
					self.splits[self.completed_laps + 1][i] = 0
			else:
				self.splits[self.completed_laps + 1][i] = current_split

		is_new_lap = self.completed_laps > self._last_lap_number or (self.current_lap and self.current_lap < 300)
		if is_new_lap and not self._lap_handled:
			self._lap_handled = True
			self._last_lap_number = self.completed_laps
			self._total_time += self.last_lap
			self.on_new_lap.emit()
		if not is_new_lap:
			self._lap_handled = False

	# -------------------------------------
	@property
	def session_key(self):
		# -------------------------------------
		return '-'.join([self.track_name_id, self.track_configuration, self.car_name])

	# -------------------------------------
	@property
	def total_time(self):
		# -------------------------------------
		return self._total_time + self.current_lap

	# -------------------------------------
	def __str__(self):
		# -------------------------------------
		return 'id: {}, laps: {}, pos: {:.5f}, current: {}, last: {}, best: {}'.format(self.car_id, self.completed_laps, self.position, self.current_lap, self.last_lap, self.best_lap)
