# !/usr/bin/env python
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# Title       : cars.py
# Author      : aDd1kTeD2Ka0s (add1kted2ka0s@gmail.com)
# Date        : 20-Jun-2014
# Description : Ka0s Leaderboard - Cars Data
#             : Original code by Rombik
# -----------------------------------------------------------------------------

# -------------------------------------
# Imports
# -------------------------------------

import ac
import itertools

from collections import namedtuple
from functools import partial

from .config import *
from .laps import Laps
from .car_state import CarState
from ..lib.sim_info import info
from ..lib.utils import log

# -------------------------------------
# Globals
# -------------------------------------

counter = itertools.count()
Deltas = namedtuple('Deltas', 'delta_time_lead, delta_time_prev, delta_time_you, delta_distance_lead_mi, delta_distance_prev_mi, delta_distance_you_mi, delta_distance_lead_km, delta_distance_prev_km, delta_distance_you_km')

# -------------------------------------
class Cars(object):
	# -------------------------------------

	# -------------------------------------
	def __init__(self):
		# -------------------------------------
		self.laps = None
		self.car_states = None

		self.own_grid_pos = None
		self.full_grid = None
		self.visible_grid = None

	# -------------------------------------
	def _init_data(self):
		# -------------------------------------
		num_cars = info.static.numCars
		self.laps = [Laps() for _ in range(num_cars)]
		self.car_states = [CarState(i) for i in range(num_cars)]
		self._restart_handled = False
		for lap, car_state, in zip(self.laps, self.car_states):
			car_state.on_new_lap.connect(partial(lap.on_new_lap, car_state))

	# -------------------------------------
	def get_players(self):
		# -------------------------------------
		result = {}
		for i in range(1000):
			name = ac.getDriverName(i)
			if isinstance(name, str):
				result[i] = name
		return result

	# -------------------------------------
	def reset(self):
		# -------------------------------------
		self._restart_handled = False
		for car_state in self.car_states:
			car_state.reset()
		for lap in self.laps:
			lap.clear()

	# -------------------------------------
	def on_update(self, dt, include_leader=True, limit=4):
		# -------------------------------------
		# TODO: or len(self.car_states) != info.static.numCars:
		# TODO: Retired car has last lap time 0?
		if not self.car_states:
			self._init_data()

		if info.graphics.sessionTimeLeft == float('inf'):
			if not self._restart_handled:
				self.reset()
				self._restart_handled = True
		else:
			self._restart_handled = False

		for car_id, car_state in enumerate(self.car_states):
			car_state.on_update(dt)
			time = car_state.total_time
			self.laps[car_id].add(car_state.completed_laps, car_state.position, time)

		self.own_grid_pos, self.full_grid = self._build_full_grid()
		self.visible_grid = self._build_visible_grid(self.own_grid_pos, self.full_grid, include_leader=include_leader, limit=limit)

	# -------------------------------------
	def _build_full_grid(self):
		# -------------------------------------
		ordered_car_states = sorted(enumerate(self.car_states), key=_grid_sort_key)
		own_grid_pos = 0
		grid_positions = []
		for grid_pos, (car, status) in enumerate(ordered_car_states):
			if car == 0:
				own_grid_pos = grid_pos
			grid_positions.append((grid_pos, car))
		return own_grid_pos, grid_positions

	# -------------------------------------
	def _build_visible_grid(self, own_grid_pos, grid_positions, include_leader=True, limit=4):
		# -------------------------------------
		# TODO: brain dump, fix this mess
		if len(grid_positions) <= limit:
			return grid_positions
		half = limit // 2
		left_bound = max(own_grid_pos - half, 0)
		right_bound = min(own_grid_pos + half, len(grid_positions))
		before = grid_positions[left_bound:own_grid_pos]
		after = grid_positions[own_grid_pos:right_bound]
		total_len = len(before) + len(after)
		free = limit - total_len
		need_to_insert_leader = include_leader and own_grid_pos != 0 and left_bound != 0
		if need_to_insert_leader:
			if not free:
				if len(before) > len(after) - 1:
					before.pop(0)
				else:
					after.pop()
			else:
				free -= 1
		if free:
			if right_bound < len(grid_positions):
				after.extend(grid_positions[right_bound:right_bound + free])
			else:
				before = grid_positions[left_bound - free:left_bound] + before
		if need_to_insert_leader:
			before.insert(0, grid_positions[0])
		result = before + after
		return result

	# -------------------------------------
	def calc_best_worst_leaderboard(self):
		# -------------------------------------
		best_worst = dict()

		# Init best/worst
		# --- Splits
		best_worst['splits'] = dict()
		best_worst['splits']['best'] = list()
		best_worst['splits']['worst'] = list()
		for i in range(SPLIT_COUNT):
			best_worst['splits']['best'].append({'car_id': None, 'value': None})
			best_worst['splits']['worst'].append({'car_id': None, 'value': None})
		# --- Last lap
		best_worst['last_lap'] = dict()
		best_worst['last_lap']['best'] = {'car_id': None, 'value': None}
		best_worst['last_lap']['worst'] = {'car_id': None, 'value': None}
		# --- Best lap
		best_worst['best_lap'] = dict()
		best_worst['best_lap']['best'] = {'car_id': None, 'value': None}
		best_worst['best_lap']['worst'] = {'car_id': None, 'value': None}
		# --- Speed max
		best_worst['speed_max'] = dict()
		best_worst['speed_max']['best'] = {'car_id': None, 'value': None}
		best_worst['speed_max']['worst'] = {'car_id': None, 'value': None}

		for grid_pos, car_id in self.visible_grid:
			car_state = self.car_states[car_id]
			# Calc best/worst splits
			for split_id in range(SPLIT_COUNT):
				if car_state.inprogress_split != split_id and car_state.live_splits[split_id] != 0 and (best_worst['splits']['best'][split_id]['value'] is None or car_state.live_splits[split_id] < best_worst['splits']['best'][split_id]['value']):
					best_worst['splits']['best'][split_id] = {'car_id': car_id, 'value': car_state.live_splits[split_id]}
				if car_state.inprogress_split != split_id and car_state.live_splits[split_id] != 0 and (best_worst['splits']['worst'][split_id]['value'] is None or car_state.live_splits[split_id] > best_worst['splits']['worst'][split_id]['value']):
					best_worst['splits']['worst'][split_id] = {'car_id': car_id, 'value': car_state.live_splits[split_id]}
			# Calc best/worst last lap
			if car_state.last_lap != 0 and (best_worst['last_lap']['best']['value'] is None or car_state.last_lap < best_worst['last_lap']['best']['value']):
				best_worst['last_lap']['best'] = {'car_id': car_id, 'value': car_state.last_lap}
			if car_state.last_lap != 0 and (best_worst['last_lap']['worst']['value'] is None or car_state.last_lap > best_worst['last_lap']['worst']['value']):
				best_worst['last_lap']['worst'] = {'car_id': car_id, 'value': car_state.last_lap}
			# Calc best/worst best lap
			if car_state.best_lap != 0 and (best_worst['best_lap']['best']['value'] is None or car_state.best_lap < best_worst['best_lap']['best']['value']):
				best_worst['best_lap']['best'] = {'car_id': car_id, 'value': car_state.best_lap}
			if car_state.best_lap != 0 and (best_worst['best_lap']['worst']['value'] is None or car_state.best_lap > best_worst['best_lap']['worst']['value']):
				best_worst['best_lap']['worst'] = {'car_id': car_id, 'value': car_state.best_lap}
			# Calc best/worst speed max
			if best_worst['speed_max']['best']['value'] is None or car_state.speed_kmh_max > best_worst['speed_max']['best']['value']:
				best_worst['speed_max']['best'] = {'car_id': car_id, 'value': car_state.speed_kmh_max}
			if best_worst['speed_max']['worst']['value'] is None or car_state.speed_kmh_max < best_worst['speed_max']['worst']['value']:
				best_worst['speed_max']['worst'] = {'car_id': car_id, 'value': car_state.speed_kmh_max}

		return best_worst

	# -------------------------------------
	def calc_best_worst_laps(self):
		# -------------------------------------
		best_worst = dict()

		# Init best/worst
		for car_state in self.car_states:
			car_id = car_state.car_id
			best_worst[car_id] = dict()
			# --- Init Splits
			best_worst[car_id]['splits'] = dict()
			best_worst[car_id]['splits']['best'] = list()
			best_worst[car_id]['splits']['worst'] = list()
			for i in range(SPLIT_COUNT):
				best_worst[car_id]['splits']['best'].append({'lap_number': None, 'value': None})
				best_worst[car_id]['splits']['worst'].append({'lap_number': None, 'value': None})
			# --- Init Lap Time
			best_worst[car_id]['lap_time'] = dict()
			best_worst[car_id]['lap_time']['best'] = {'lap_number': None, 'value': None}
			best_worst[car_id]['lap_time']['worst'] = {'lap_number': None, 'value': None}
			# --- Loop through laps
			for lap_number, lap in car_state.laps.items():
				if (car_state.completed_laps + 1) != lap_number and (best_worst[car_id]['lap_time']['best']['value'] is None or lap < best_worst[car_id]['lap_time']['best']['value']):
					best_worst[car_id]['lap_time']['best'] = {'lap_number': lap_number, 'value': lap}
				if (car_state.completed_laps + 1) != lap_number and (best_worst[car_id]['lap_time']['worst']['value'] is None or lap > best_worst[car_id]['lap_time']['worst']['value']):
					best_worst[car_id]['lap_time']['worst'] = {'lap_number': lap_number, 'value': lap}

			for lap_number, splits in car_state.splits.items():
				for split_id in range(SPLIT_COUNT):
					if ((car_state.completed_laps + 1) != lap_number or ((car_state.completed_laps + 1) == lap_number and car_state.inprogress_split != split_id and splits[split_id] != 0)) and (best_worst[car_id]['splits']['best'][split_id]['value'] is None or splits[split_id] < best_worst[car_id]['splits']['best'][split_id]['value']):
						best_worst[car_id]['splits']['best'][split_id] = {'lap_number': lap_number, 'value': splits[split_id]}
					if ((car_state.completed_laps + 1) != lap_number or ((car_state.completed_laps + 1) == lap_number and car_state.inprogress_split != split_id and splits[split_id] != 0)) and (best_worst[car_id]['splits']['worst'][split_id]['value'] is None or splits[split_id] > best_worst[car_id]['splits']['worst'][split_id]['value']):
						best_worst[car_id]['splits']['worst'][split_id] = {'lap_number': lap_number, 'value': splits[split_id]}
		return best_worst

	# -------------------------------------
	def calc_deltas(self):
		# -------------------------------------
		entry = dict()
		deltas = list()

		# Calculate metadata for YOU
		id_you = 0
		grid_pos_you = self.own_grid_pos
		self.calc_deltas_build_entry(entry, 'YOU', id_you, grid_pos_you)

		# Calculate metadata for LEAD
		id_lead = self.visible_grid[0][1]
		grid_pos_lead = 0
		self.calc_deltas_build_entry(entry, 'LEAD', id_lead, grid_pos_lead)

		for grid_pos, car_id in self.visible_grid:
			car_state = self.car_states[car_id]
			# Calculate metadata for PREV
			if grid_pos > 0:
				if (grid_pos - 1) not in self.visible_grid:
					id_prev = self.full_grid[grid_pos - 1][1]
				else:
					id_prev = self.visible_grid[grid_pos - 1][1]
				grid_pos_prev = grid_pos - 1
			else:
				id_prev = id_lead
				grid_pos_prev = grid_pos_lead
			self.calc_deltas_build_entry(entry, 'PREV', id_prev, grid_pos_prev)

			# Calculate deltas
			delta_time_lead, delta_distance_lead_km, delta_distance_lead_mi = self.calc_deltas_calculation(car_id, grid_pos, entry['LEAD'])
			delta_time_prev, delta_distance_prev_km, delta_distance_prev_mi = self.calc_deltas_calculation(car_id, grid_pos, entry['PREV'])
			delta_time_you, delta_distance_you_km, delta_distance_you_mi = self.calc_deltas_calculation(car_id, grid_pos, entry['YOU'])

			# Build deltas object
			deltas.append(Deltas(delta_time_lead, delta_time_prev, delta_time_you, delta_distance_lead_mi, delta_distance_prev_mi, delta_distance_you_mi, delta_distance_lead_km, delta_distance_prev_km, delta_distance_you_km))

		return deltas

	# -------------------------------------
	def calc_deltas_build_entry(self, entry, mode, car_id, grid_pos):
		# -------------------------------------
		entry[mode] = dict()
		entry[mode]['grid_pos'] = grid_pos
		entry[mode]['laps'] = self.laps[car_id]
		entry[mode]['completed_laps'] = self.car_states[car_id].completed_laps
		entry[mode]['position'] = self.car_states[car_id].position
		entry[mode]['time'] = self.car_states[car_id].total_time
		entry[mode]['distance_km'] = self.car_states[car_id].distance_race_km
		entry[mode]['distance_mi'] = self.car_states[car_id].distance_race_mi

	# -------------------------------------
	def calc_deltas_calculation(self, car_id, grid_pos, entry):
		# -------------------------------------
		# Get sign and time at position
		if grid_pos > entry['grid_pos']:
			time_at_position = entry['laps'].time_at_position(self.car_states[car_id].completed_laps, self.car_states[car_id].position)
			sign_lead = 1
		else:
			time_at_position = self.laps[car_id].time_at_position(entry['completed_laps'], entry['position'])
			sign_lead = -1

		# Calculate time/distance delta
		if time_at_position is None:
			delta_time = None
		else:
			delta_time = (entry['time'] - time_at_position) / 1000.0 * sign_lead
		delta_km = abs(entry['distance_km'] - self.car_states[car_id].distance_race_km) * sign_lead
		delta_mi = abs(entry['distance_mi'] - self.car_states[car_id].distance_race_mi) * sign_lead

		return (delta_time, delta_km, delta_mi)

# -------------------------------------
def _grid_sort_key(e):
	# -------------------------------------
	rs = e[1]
	return -rs.completion_race
