# !/usr/bin/env python
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# Title       : laps.py
# Author      : aDd1kTeD2Ka0s (add1kted2ka0s@gmail.com)
# Date        : 20-Jun-2014
# Description : Ka0s Leaderboard - Lap Data
#             : Original code by Rombik
# -----------------------------------------------------------------------------

# -------------------------------------
# Imports
# -------------------------------------

import array
import bisect
import struct

from ..lib.utils import log

# -------------------------------------
def interpolate(x, y0, y1, x0, x1):
	# -------------------------------------
	# Linear interpolation

	return y0 + (y1 - y0) * float(x - x0) / (x1 - x0)

# -------------------------------------
class Lap(object):
	# -------------------------------------

	# : File format version
	VERSION = 1

	# -------------------------------------
	def __init__(self, resolution_ms=10):
		# -------------------------------------
		#:param int resolution_ms:
		#	Data will be saved only not often than once in this interval,
		#	e.g. if `resolution_ms` is 100, then 1000/100 = 10 records
		#	per second will be saved, or about 4*2*10 = 80 bytes/second
		#	default is 10, for more precise data. If you use it for online leaderboard,
		#	consider greater values.

		self.resolution_ms = resolution_ms
		self.positions = array.array('f')
		self.times = array.array('I')
		self._header_struct = struct.Struct('II')

	# -------------------------------------
	def add(self, position, time):
		# -------------------------------------
		self._fix_points_for_wrong_spline()
		if not self._is_need_to_skip(position, time):
			self.positions.append(position)
			self.times.append(time)

	# -------------------------------------
	def _fix_points_for_wrong_spline(self):
		# -------------------------------------
		# Handles case, when first point has position like 0.9,
		# It happens when spline has wrong start point, e.g. from AI Recorder app

		while self.positions and self.positions[0] > 0.5:
			# yes, it's not effective, but there shouldn't be many points, so it's ok
			self.positions.pop(0)
			self.times.pop(0)

	# -------------------------------------
	def _is_need_to_skip(self, position, time):
		# -------------------------------------
		if self.positions:
			# skip bakwards rides
			if self.positions[-1] > position:
				return True
			# rate limit
			if time - self.times[-1] < self.resolution_ms:
				return True
		return False

	# -------------------------------------
	def time_at_position(self, position, interpolate=interpolate):
		# -------------------------------------
		times, positions = self.times, self.positions
		i = bisect.bisect_right(positions, position) - 1
		if i + 1 >= len(times) or i < 0:
			try:
				return times[i]
			except IndexError:
				return None
		interpolated_time = interpolate(position, times[i], times[i + 1], positions[i], positions[i + 1])
		return interpolated_time

	# -------------------------------------
	def clear(self):
		# -------------------------------------
		del self.positions[:]
		del self.times[:]

	# -------------------------------------
	def __len__(self):
		# -------------------------------------
		return len(self.positions)

# -------------------------------------
class Laps(object):
	# -------------------------------------

	# -------------------------------------
	def __init__(self, max_laps=3, resolution_ms=10):
		# -------------------------------------
		self.laps = {}
		self.max_laps = max_laps
		self.resolution_ms = resolution_ms

	# -------------------------------------
	def add(self, lap, position, time):
		# -------------------------------------
		if lap not in self.laps:
			self.laps[lap] = Lap(resolution_ms=10)
			if len(self.laps) > self.max_laps:
				min_lap_number = min(self.laps)
				self.laps.pop(min_lap_number)
		self.laps[lap].add(position, time)

	# -------------------------------------
	def clear(self):
		# -------------------------------------
		self.laps.clear()

	# -------------------------------------
	def time_at_position(self, lap, position, default=None):
		# -------------------------------------
		if lap not in self.laps:
			return default
		return self.laps[lap].time_at_position(position)

	# -------------------------------------
	def __len__(self):
		# -------------------------------------
		return sum(len(ld) for ld in self.laps.values())

	# -------------------------------------
	def on_new_lap(self, car_state):
		# -------------------------------------
		pass
